from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestRange(BoaTest):
    default_folder: str = 'test_sc/range_test'

    RANGE_ERROR_MESSAGE = String('range() arg 3 must not be zero').to_bytes()

    def test_range_given_length(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.PUSH1      # range(arg0)
            + Opcode.PUSH0
            + Opcode.LDARG0
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIGN
            + Opcode.JMPIF
            + Integer(5 + len(self.RANGE_ERROR_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1
            + Integer(len(self.RANGE_ERROR_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + self.RANGE_ERROR_MESSAGE
            + Opcode.THROW
            + Opcode.NEWARRAY0
            + Opcode.REVERSE4
            + Opcode.SWAP
            + Opcode.JMP
            + Integer(8).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.OVER
            + Opcode.APPEND
            + Opcode.OVER
            + Opcode.ADD
            + Opcode.DUP
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.SIGN
            + Opcode.PUSH0
            + Opcode.JMPGT
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.GT
            + Opcode.JMP
            + Integer(3).to_byte_array(signed=True, min_length=1)
            + Opcode.LT
            + Opcode.JMPIF
            + Integer(-19).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.RET        # return
        )

        path = self.get_contract_path('RangeGivenLen.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'range_example', 5)
        self.assertEqual([0, 1, 2, 3, 4], result)
        result = self.run_smart_contract(engine, path, 'range_example', 10)
        self.assertEqual([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], result)
        result = self.run_smart_contract(engine, path, 'range_example', 0)
        self.assertEqual([], result)

    def test_range_given_start(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.PUSH1      # range(arg0, arg1)
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIGN
            + Opcode.JMPIF
            + Integer(5 + len(self.RANGE_ERROR_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1
            + Integer(len(self.RANGE_ERROR_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + self.RANGE_ERROR_MESSAGE
            + Opcode.THROW
            + Opcode.NEWARRAY0
            + Opcode.REVERSE4
            + Opcode.SWAP
            + Opcode.JMP
            + Integer(8).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.OVER
            + Opcode.APPEND
            + Opcode.OVER
            + Opcode.ADD
            + Opcode.DUP
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.SIGN
            + Opcode.PUSH0
            + Opcode.JMPGT
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.GT
            + Opcode.JMP
            + Integer(3).to_byte_array(signed=True, min_length=1)
            + Opcode.LT
            + Opcode.JMPIF
            + Integer(-19).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.RET        # return
        )

        path = self.get_contract_path('RangeGivenStart.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'range_example', 2, 6)
        self.assertEqual([2, 3, 4, 5], result)
        result = self.run_smart_contract(engine, path, 'range_example', -10, 0)
        self.assertEqual([-10, -9, -8, -7, -6, -5, -4, -3, -2, -1], result)

    def test_range_given_step(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x03'
            + Opcode.LDARG2     # range(arg0, arg1, arg2)
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIGN
            + Opcode.JMPIF
            + Integer(5 + len(self.RANGE_ERROR_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1
            + Integer(len(self.RANGE_ERROR_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + self.RANGE_ERROR_MESSAGE
            + Opcode.THROW
            + Opcode.NEWARRAY0
            + Opcode.REVERSE4
            + Opcode.SWAP
            + Opcode.JMP
            + Integer(8).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.OVER
            + Opcode.APPEND
            + Opcode.OVER
            + Opcode.ADD
            + Opcode.DUP
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.SIGN
            + Opcode.PUSH0
            + Opcode.JMPGT
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.GT
            + Opcode.JMP
            + Integer(3).to_byte_array(signed=True, min_length=1)
            + Opcode.LT
            + Opcode.JMPIF
            + Integer(-19).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.RET        # return
        )

        path = self.get_contract_path('RangeGivenStep.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'range_example', 2, 10, 3)
        self.assertEqual([2, 5, 8], result)
        result = self.run_smart_contract(engine, path, 'range_example', -2, 10, 3)
        self.assertEqual([-2, 1, 4, 7], result)

    def test_range_parameter_mismatched_type(self):
        path = self.get_contract_path('RangeParameterMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_range_as_sequence(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.PUSH1      # range(arg0, arg1)
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIGN
            + Opcode.JMPIF
            + Integer(5 + len(self.RANGE_ERROR_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHDATA1
            + Integer(len(self.RANGE_ERROR_MESSAGE)).to_byte_array(signed=True, min_length=1)
            + self.RANGE_ERROR_MESSAGE
            + Opcode.THROW
            + Opcode.NEWARRAY0
            + Opcode.REVERSE4
            + Opcode.SWAP
            + Opcode.JMP
            + Integer(8).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.OVER
            + Opcode.APPEND
            + Opcode.OVER
            + Opcode.ADD
            + Opcode.DUP
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.PUSH3
            + Opcode.PICK
            + Opcode.SIGN
            + Opcode.PUSH0
            + Opcode.JMPGT
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.GT
            + Opcode.JMP
            + Integer(3).to_byte_array(signed=True, min_length=1)
            + Opcode.LT
            + Opcode.JMPIF
            + Integer(-19).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.RET        # return
        )

        path = self.get_contract_path('RangeExpectedSequence.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'range_example', 2, 6)
        self.assertEqual([2, 3, 4, 5], result)
        result = self.run_smart_contract(engine, path, 'range_example', -10, 0)
        self.assertEqual([-10, -9, -8, -7, -6, -5, -4, -3, -2, -1], result)

    def test_range_mismatched_type(self):
        path = self.get_contract_path('RangeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_range_too_few_parameters(self):
        path = self.get_contract_path('RangeTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_range_too_many_parameters(self):
        path = self.get_contract_path('RangeTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_range_get_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('GetValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', [1, 2, 3, 4])
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', [5, 3, 2])
        self.assertEqual(5, result)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX):
            self.run_smart_contract(engine, path, 'Main', [])

    def test_range_set_value(self):
        path = self.get_contract_path('SetValue.py')
        self.assertCompilerLogs(CompilerError.UnresolvedOperation, path)

    def test_range_slicing(self):
        path = self.get_contract_path('RangeSlicingLiteralValues.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([2], result)

    def test_range_slicing_start_larger_than_ending(self):
        path = self.get_contract_path('RangeSlicingStartLargerThanEnding.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_range_slicing_with_variables(self):
        path = self.get_contract_path('RangeSlicingVariableValues.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([2], result)

    def test_range_slicing_negative_start(self):
        path = self.get_contract_path('RangeSlicingNegativeStart.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([2, 3, 4, 5], result)

    def test_range_slicing_negative_end(self):
        path = self.get_contract_path('RangeSlicingNegativeEnd.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([0, 1], result)

    def test_range_slicing_start_omitted(self):
        path = self.get_contract_path('RangeSlicingStartOmitted.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([0, 1, 2], result)

    def test_range_slicing_omitted(self):
        path = self.get_contract_path('RangeSlicingOmitted.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([0, 1, 2, 3, 4, 5], result)

    def test_range_slicing_end_omitted(self):
        path = self.get_contract_path('RangeSlicingEndOmitted.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([2, 3, 4, 5], result)

    def test_range_slicing_with_stride(self):
        path = self.get_contract_path('RangeSlicingWithStride.py')
        engine = TestEngine()

        a = range(6)
        expected_result = a[2:5:2]
        result = self.run_smart_contract(engine, path, 'literal_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[2:5:2]
        result = self.run_smart_contract(engine, path, 'literal_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-6:5:2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[0:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-6:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-999:5:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[0:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-999:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[999:5:2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[0:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[999:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_values')
        self.assertEqual(list(expected_result), result)

    def test_range_slicing_with_negative_stride(self):
        path = self.get_contract_path('RangeSlicingWithNegativeStride.py')
        engine = TestEngine()

        a = range(6)
        expected_result = a[2:5:-1]
        result = self.run_smart_contract(engine, path, 'literal_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-6:5:-1]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[0:-1:-1]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-6:-1:-1]
        result = self.run_smart_contract(engine, path, 'negative_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-999:5:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[0:-999:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-999:-999:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[999:5:-1]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[0:999:-1]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[999:999:-1]
        result = self.run_smart_contract(engine, path, 'really_high_values')
        self.assertEqual(list(expected_result), result)

    def test_range_slicing_omitted_with_stride(self):
        path = self.get_contract_path('RangeSlicingOmittedWithStride.py')
        engine = TestEngine()

        a = range(6)
        expected_result = a[::2]
        result = self.run_smart_contract(engine, path, 'omitted_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[:5:2]
        result = self.run_smart_contract(engine, path, 'omitted_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[2::2]
        result = self.run_smart_contract(engine, path, 'omitted_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-6::2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-999::2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[999::2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(list(expected_result), result)

    def test_range_slicing_omitted_with_negative_stride(self):
        path = self.get_contract_path('RangeSlicingOmittedWithNegativeStride.py')
        engine = TestEngine()

        a = range(6)
        expected_result = a[::-2]
        result = self.run_smart_contract(engine, path, 'omitted_values')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[:5:-2]
        result = self.run_smart_contract(engine, path, 'omitted_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[2::-2]
        result = self.run_smart_contract(engine, path, 'omitted_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-6::-2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[:-1:-2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[-999::-2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[:-999:-2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[999::-2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(list(expected_result), result)

        a = range(6)
        expected_result = a[:999:-2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(list(expected_result), result)

    def test_boa2_range_test(self):
        path = self.get_contract_path('RangeBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(list(range(100, 120)), result)

    def test_range_index(self):
        path = self.get_contract_path('IndexRange.py')
        engine = TestEngine()

        # TODO: change when index() with only one argument is implemented for range
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)



from typing import Dict

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.interop.storage.storagecontext.storagecontexttype import StorageContextType
from boa3.model.variable import Variable


class StorageGetReadOnlyContextMethod(InteropMethod):

    def __init__(self, storage_context_type: StorageContextType):
        identifier = 'get_read_only_context'
        native_identifier = 'System.Storage.GetContext'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, native_identifier, args, return_type=storage_context_type)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo3.contracts import FindOptions
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestRelational(BoaTest):
    default_folder: str = 'test_sc/relational_test'

    # region GreaterThan

    def test_builtin_type_greater_than_operation(self):
        path = self.get_contract_path('BuiltinTypeGreaterThan.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES)
        self.assertEqual(FindOptions.VALUES_ONLY > FindOptions.DESERIALIZE_VALUES, result)

    def test_mixed_greater_than_operation(self):
        path = self.get_contract_path('MixedGreaterThan.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_number_greater_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.GT
            + Opcode.RET
        )

        path = self.get_contract_path('NumGreaterThan.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 2)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 2)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 1)
        self.assertEqual(True, result)

    def test_string_greater_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.GT
            + Opcode.RET
        )

        path = self.get_contract_path('StrGreaterThan.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'test', 'unit')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'unit')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'test')
        self.assertEqual(False, result)

    # endregion

    # region GreaterThanOrEqual

    def test_builtin_type_greater_than_or_equal_operation(self):
        path = self.get_contract_path('BuiltinTypeGreaterThanOrEqual.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.VALUES_ONLY, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.VALUES_ONLY >= FindOptions.VALUES_ONLY, result)

    def test_mixed_greater_or_equal_than_operation(self):
        path = self.get_contract_path('MixedGreaterOrEqual.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_number_greater_or_equal_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.GE
            + Opcode.RET
        )

        path = self.get_contract_path('NumGreaterOrEqual.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 2)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 2)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 1)
        self.assertEqual(True, result)

    def test_string_greater_or_equal_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.GE
            + Opcode.RET
        )

        path = self.get_contract_path('StrGreaterOrEqual.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'test', 'unit')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'unit')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'test')
        self.assertEqual(False, result)

    # endregion

    # region Identity

    def test_boolean_identity_operation(self):
        path = self.get_contract_path('BoolIdentity.py')
        engine = TestEngine()

        a = True
        b = True
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'without_attribution_true')
        self.assertEqual(expected_result, result)

        a = True
        b = False
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'without_attribution_false')
        self.assertEqual(expected_result, result)

        c = True
        d = c
        expected_result = c is d
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

    def test_list_identity(self):
        path = self.get_contract_path('ListIdentity.py')
        engine = TestEngine()

        a = [1, 2, 3]
        b = a
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

        a = [1, 2, 3]
        b = [1, 2, 3]
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'without_attribution')
        self.assertEqual(expected_result, result)

    def test_mixed_identity(self):
        path = self.get_contract_path('MixedIdentity.py')
        engine = TestEngine()

        # a mixed identity should always result in False, but will compile
        result = self.run_smart_contract(engine, path, 'mixed')
        self.assertEqual(False, result)

    def test_none_identity_operation(self):
        path = self.get_contract_path('NoneIdentity.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', True)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 'string')
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', b'bytes')
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', None)
        self.assertEqual(True, result)

    def test_number_identity_operation(self):
        path = self.get_contract_path('NumIdentity.py')
        engine = TestEngine()

        a = 1
        b = 1
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'without_attribution_true')
        self.assertEqual(expected_result, result)

        a = 1
        b = 2
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'without_attribution_false')
        self.assertEqual(expected_result, result)

        c = 1
        d = c
        expected_result = c is d
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

    def test_string_identity_operation(self):
        path = self.get_contract_path('StrIdentity.py')
        engine = TestEngine()

        a = 'unit'
        b = 'unit'
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'without_attribution_true')
        self.assertEqual(expected_result, result)

        a = 'unit'
        b = 'test'
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'without_attribution_false')
        self.assertEqual(expected_result, result)

        c = 'unit'
        d = c
        expected_result = c is d
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

    def test_tuple_identity(self):
        path = self.get_contract_path('TupleIdentity.py')
        engine = TestEngine()

        a = (1, 2, 3)
        b = a
        expected_result = a is b
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

        # Python will try conserve memory and will make a and b reference the same position, since Tuples are immutable
        # this will deviate from Neo's expected behavior
        result = self.run_smart_contract(engine, path, 'without_attribution')
        self.assertEqual(False, result)

    # endregion

    # region LessThan

    def test_builtin_type_less_than_operation(self):
        path = self.get_contract_path('BuiltinTypeLessThan.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.VALUES_ONLY, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.VALUES_ONLY < FindOptions.VALUES_ONLY, result)

    def test_mixed_less_than_operation(self):
        path = self.get_contract_path('MixedLessThan.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_number_less_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.LT
            + Opcode.RET
        )

        path = self.get_contract_path('NumLessThan.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 2)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 2)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 1)
        self.assertEqual(False, result)

    def test_string_less_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.LT
            + Opcode.RET
        )

        path = self.get_contract_path('StrLessThan.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'test', 'unit')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'unit')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'test')
        self.assertEqual(True, result)

    # endregion

    # region LessThanOrEqual

    def test_builtin_type_less_than_or_equal_operation(self):
        path = self.get_contract_path('BuiltinTypeLessThanOrEqual.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.VALUES_ONLY, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.VALUES_ONLY <= FindOptions.VALUES_ONLY, result)

    def test_mixed_less_or_equal_than_operation(self):
        path = self.get_contract_path('MixedLessOrEqual.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_number_less_or_equal_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.LE
            + Opcode.RET
        )

        path = self.get_contract_path('NumLessOrEqual.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 2)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 2)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 1)
        self.assertEqual(False, result)

    def test_string_less_or_equal_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.LE
            + Opcode.RET
        )

        path = self.get_contract_path('StrLessOrEqual.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'test', 'unit')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'unit')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'test')
        self.assertEqual(True, result)

    # endregion

    # region MixedEquality

    def test_mixed_equality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.EQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('MixedEquality.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 'unit')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 123, '123')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', Integer.from_bytes(b'123'), '123')
        self.assertEqual(False, result)

    def test_boa2_equality_test2(self):
        path = self.get_contract_path('Equality2Boa2Test.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 2)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 4)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 5)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 6)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 7)
        self.assertEqual(False, result)

    # endregion

    # region MixedInequality

    def test_mixed_inequality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NOTEQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('MixedInequality.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 'unit')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 123, '123')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', Integer.from_bytes(b'123'), '123')
        self.assertEqual(True, result)

    # endregion

    # region NotIdentity

    def test_boolean_not_identity_operation(self):
        path = self.get_contract_path('BoolNotIdentity.py')
        engine = TestEngine()

        a = True
        b = False
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'without_attribution_true')
        self.assertEqual(expected_result, result)

        a = True
        b = True
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'without_attribution_false')
        self.assertEqual(expected_result, result)

        c = True
        d = c
        expected_result = c is not d
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

    def test_list_not_identity(self):
        path = self.get_contract_path('ListNotIdentity.py')
        engine = TestEngine()

        a = [1, 2, 3]
        b = a
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

        a = [1, 2, 3]
        b = [1, 2, 3]
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'without_attribution')
        self.assertEqual(expected_result, result)

    def test_none_not_identity_operation(self):
        path = self.get_contract_path('NoneNotIdentity.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', True)
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 'string')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', b'bytes')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', None)
        self.assertEqual(False, result)

    def test_number_not_identity_operation(self):
        path = self.get_contract_path('NumNotIdentity.py')
        engine = TestEngine()

        a = 1
        b = 2
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'without_attribution_true')
        self.assertEqual(expected_result, result)

        a = 1
        b = 1
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'without_attribution_false')
        self.assertEqual(expected_result, result)

        c = 1
        d = c
        expected_result = c is not d
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

    def test_string_not_identity_operation(self):
        path = self.get_contract_path('StrNotIdentity.py')
        engine = TestEngine()

        a = 'unit'
        b = 'test'
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'without_attribution_true')
        self.assertEqual(expected_result, result)

        a = 'unit'
        b = 'unit'
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'without_attribution_false')
        self.assertEqual(expected_result, result)

        c = 'unit'
        d = c
        expected_result = c is not d
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

    def test_tuple_not_identity(self):
        path = self.get_contract_path('TupleNotIdentity.py')
        engine = TestEngine()

        a = (1, 2, 3)
        b = a
        expected_result = a is not b
        result = self.run_smart_contract(engine, path, 'with_attribution')
        self.assertEqual(expected_result, result)

        # Python will try conserve memory and will make a and b reference the same position, since Tuples are immutable
        # this will deviate from Neo's expected behavior
        result = self.run_smart_contract(engine, path, 'without_attribution')
        self.assertEqual(True, result)

    # endregion

    # region NumericEquality

    def test_boolean_equality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NUMEQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('BoolEquality.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', True, True)
        self.assertEqual(True, result)

    def test_builtin_equality_operation(self):
        path = self.get_contract_path('BuiltinTypeEquality.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.VALUES_ONLY, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.VALUES_ONLY == FindOptions.VALUES_ONLY, result)

    def test_multiple_comparisons(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x03'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.LE
            + Opcode.LDARG0
            + Opcode.LDARG2
            + Opcode.LE
            + Opcode.BOOLAND
            + Opcode.RET
        )

        path = self.get_contract_path('NumRange.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 2, 5)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 1, 5)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 5, 1, 2)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 5, 1)
        self.assertEqual(False, result)

    def test_number_equality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NUMEQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('NumEquality.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 2)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 2)
        self.assertEqual(True, result)

    # endregion

    # region NumericInequality

    def test_boolean_inequality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NUMNOTEQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('BoolInequality.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', True, True)
        self.assertEqual(False, result)

    def test_builtin_inequality_operation(self):
        path = self.get_contract_path('BuiltinTypeInequality.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.VALUES_ONLY, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.VALUES_ONLY != FindOptions.VALUES_ONLY, result)

    def test_number_inequality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NUMNOTEQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('NumInequality.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1, 2)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 2, 2)
        self.assertEqual(False, result)

    def test_number_inequality_operation_2(self):
        path = self.get_contract_path('NumInequalityPython2.py')

        with self.assertRaises(SyntaxError):
            Boa3.compile(path)

    # endregion

    # region ObjectEquality

    def test_compare_same_value_argument(self):
        path = self.get_contract_path('CompareSameValueArgument.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'testing_something', bytes(20))
        self.assertEqual(True, result)

    def test_compare_same_value_hard_coded(self):
        path = self.get_contract_path('CompareSameValueHardCoded.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'testing_something')
        self.assertEqual(True, result)

    def test_compare_string(self):
        path = self.get_contract_path('CompareString.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test1', '|')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'test2', '|')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'test3', '|')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'test4', '|')
        self.assertEqual(True, result)

    def test_list_equality_with_slice(self):
        path = self.get_contract_path('ListEqualityWithSlice.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', ['unittest', '123'], 'unittest')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', ['unittest', '123'], '123')
        self.assertEqual(False, result)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX):
            self.run_smart_contract(engine, path, 'main', [], '')

    def test_string_equality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.EQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('StrEquality.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'test')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'unit')
        self.assertEqual(True, result)

    # endregion

    # region ObjectInequality

    def test_string_inequality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NOTEQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('StrInequality.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'test')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 'unit')
        self.assertEqual(False, result)

    # endregion



import ast
from typing import Optional, Tuple, Union

from boa3.model.expression import IExpression
from boa3.model.imports.package import Package
from boa3.model.symbol import ISymbol
from boa3.model.type.classes.classtype import ClassType
from boa3.model.type.itype import IType


class Attribute(IExpression):
    """
    A class used to represent an attribute

    :ivar value: the origin expression that has the attribute
    :ivar attr_name: the name of the attribute
    :ivar attr_symbol: the found symbol for the attribute
    """

    def __init__(self, value: Union[ast.AST, IExpression, Package], attr_name: str,
                 attr_symbol: Optional[ISymbol] = None, origin: Optional[ast.AST] = None):
        super().__init__(origin)

        self.value: Union[ast.AST, IExpression, Package] = value
        self.attr_name: str = attr_name

        obj_with_symbols = value.type if isinstance(value, IExpression) else value
        if (isinstance(value, (IExpression, ClassType, Package))
                and hasattr(obj_with_symbols, 'symbols') and attr_name in obj_with_symbols.symbols):
            attr_symbol = obj_with_symbols.symbols[attr_name]

        self.attr_symbol: Optional[ISymbol] = attr_symbol

    @property
    def shadowing_name(self) -> str:
        return 'attribute'

    @property
    def type(self) -> IType:
        from boa3.model.type.type import Type
        return self.attr_symbol.type if isinstance(self.attr_symbol, IExpression) else Type.none

    @property
    def values(self) -> Tuple[Union[ast.AST, IExpression], Optional[ISymbol], str]:
        return self.value, self.attr_symbol, self.attr_name



import ast
from abc import abstractmethod
from typing import Optional

from boa3.model.expression import IExpression
from boa3.model.method import Method
from boa3.model.symbol import ISymbol


class IDecorator(Method):

    @abstractmethod
    def validate_parameters(self, *params: IExpression) -> bool:
        """
        Verifies if the given parameters are valid to the method

        :param params: arguments of the method
        :return: True if all arguments are valid. False otherwise.
        """
        pass

    def update_args(self, args: ast.arguments, origin: Optional[ISymbol] = None):
        """
        Updates the given args object if this decorator has any specific cases.
        If this is not the case, DON'T overwrite this method

        :param args: arguments of the method
        :param origin: internal symbol that is
        """
        return

    @property
    def has_cls_or_self(self) -> bool:
        return False

    @property
    def is_function_decorator(self) -> bool:
        return True

    @property
    def is_class_decorator(self) -> bool:
        return False



import json

from boa3 import constants
from boa3.exception import CompilerError
from boa3.neo.vm.type.StackItem import StackItemType, serialize
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestStdlibClass(BoaTest):
    default_folder: str = 'test_sc/native_test/stdlib'

    def test_get_hash(self):
        path = self.get_contract_path('GetHash.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(constants.STD_LIB_SCRIPT, result)

    def test_base64_encode(self):
        import base64
        path = self.get_contract_path('Base64Encode.py')
        engine = TestEngine()
        expected_result = base64.b64encode(b'unit test')
        result = self.run_smart_contract(engine, path, 'Main', b'unit test',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        expected_result = base64.b64encode(b'')
        result = self.run_smart_contract(engine, path, 'Main', b'',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        long_byte_string = (b'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam accumsan magna eu massa '
                            b'vulputate bibendum. Aliquam commodo euismod tristique. Sed purus erat, pretium ut interdum '
                            b'et, aliquet sed mauris. Curabitur vitae turpis euismod, hendrerit mi a, rhoncus justo. Mauris '
                            b'sollicitudin, nisl sit amet feugiat pharetra, odio ligula congue tellus, vel pellentesque '
                            b'libero leo id dui. Morbi vel risus vehicula, consectetur mauris eget, gravida ligula. '
                            b'Maecenas aliquam velit sit amet nisi ultricies, ac sollicitudin nisi mollis. Lorem ipsum '
                            b'dolor sit amet, consectetur adipiscing elit. Ut tincidunt, nisi in ullamcorper ornare, '
                            b'est enim dictum massa, id aliquet justo magna in purus.')
        expected_result = base64.b64encode(long_byte_string)
        result = self.run_smart_contract(engine, path, 'Main', long_byte_string,
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_base64_encode_mismatched_type(self):
        path = self.get_contract_path('Base64EncodeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_base64_decode(self):
        import base64
        path = self.get_contract_path('Base64Decode.py')
        engine = TestEngine()
        arg = String.from_bytes(base64.b64encode(b'unit test'))
        result = self.run_smart_contract(engine, path, 'Main', arg,
                                         expected_result_type=bytes)
        self.assertEqual(b'unit test', result)

        arg = String.from_bytes(base64.b64encode(b''))
        result = self.run_smart_contract(engine, path, 'Main', arg,
                                         expected_result_type=bytes)
        self.assertEqual(b'', result)

        long_string = ('Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam accumsan magna eu massa '
                       'vulputate bibendum. Aliquam commodo euismod tristique. Sed purus erat, pretium ut interdum '
                       'et, aliquet sed mauris. Curabitur vitae turpis euismod, hendrerit mi a, rhoncus justo. Mauris '
                       'sollicitudin, nisl sit amet feugiat pharetra, odio ligula congue tellus, vel pellentesque '
                       'libero leo id dui. Morbi vel risus vehicula, consectetur mauris eget, gravida ligula. '
                       'Maecenas aliquam velit sit amet nisi ultricies, ac sollicitudin nisi mollis. Lorem ipsum '
                       'dolor sit amet, consectetur adipiscing elit. Ut tincidunt, nisi in ullamcorper ornare, '
                       'est enim dictum massa, id aliquet justo magna in purus.')
        arg = String.from_bytes(base64.b64encode(String(long_string).to_bytes()))
        result = self.run_smart_contract(engine, path, 'Main', arg,
                                         expected_result_type=bytes)
        self.assertEqual(String(long_string).to_bytes(), result)

    def test_base64_decode_mismatched_type(self):
        path = self.get_contract_path('Base64DecodeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_base58_encode(self):
        import base58
        path = self.get_contract_path('Base58Encode.py')
        engine = TestEngine()
        expected_result = base58.b58encode('unit test')
        result = self.run_smart_contract(engine, path, 'Main', 'unit test',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        expected_result = base58.b58encode('')
        result = self.run_smart_contract(engine, path, 'Main', '',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        long_string = ('Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam accumsan magna eu massa '
                       'vulputate bibendum. Aliquam commodo euismod tristique. Sed purus erat, pretium ut interdum '
                       'et, aliquet sed mauris. Curabitur vitae turpis euismod, hendrerit mi a, rhoncus justo. Mauris '
                       'sollicitudin, nisl sit amet feugiat pharetra, odio ligula congue tellus, vel pellentesque '
                       'libero leo id dui. Morbi vel risus vehicula, consectetur mauris eget, gravida ligula. '
                       'Maecenas aliquam velit sit amet nisi ultricies, ac sollicitudin nisi mollis. Lorem ipsum '
                       'dolor sit amet, consectetur adipiscing elit. Ut tincidunt, nisi in ullamcorper ornare, '
                       'est enim dictum massa, id aliquet justo magna in purus.')
        expected_result = base58.b58encode(long_string)
        result = self.run_smart_contract(engine, path, 'Main', long_string,
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_base58_encode_mismatched_type(self):
        path = self.get_contract_path('Base58EncodeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_base58_decode(self):
        import base58
        path = self.get_contract_path('Base58Decode.py')
        engine = TestEngine()
        arg = base58.b58encode('unit test')
        result = self.run_smart_contract(engine, path, 'Main', arg)
        self.assertEqual('unit test', result)

        arg = base58.b58encode('')
        result = self.run_smart_contract(engine, path, 'Main', arg)
        self.assertEqual('', result)

        long_string = ('Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam accumsan magna eu massa '
                       'vulputate bibendum. Aliquam commodo euismod tristique. Sed purus erat, pretium ut interdum '
                       'et, aliquet sed mauris. Curabitur vitae turpis euismod, hendrerit mi a, rhoncus justo. Mauris '
                       'sollicitudin, nisl sit amet feugiat pharetra, odio ligula congue tellus, vel pellentesque '
                       'libero leo id dui. Morbi vel risus vehicula, consectetur mauris eget, gravida ligula. '
                       'Maecenas aliquam velit sit amet nisi ultricies, ac sollicitudin nisi mollis. Lorem ipsum '
                       'dolor sit amet, consectetur adipiscing elit. Ut tincidunt, nisi in ullamcorper ornare, '
                       'est enim dictum massa, id aliquet justo magna in purus.')
        arg = base58.b58encode(long_string)
        result = self.run_smart_contract(engine, path, 'Main', arg)
        self.assertEqual(long_string, result)

    def test_base58_decode_mismatched_type(self):
        path = self.get_contract_path('Base58DecodeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_base58_check_decode(self):
        import base58
        path = self.get_contract_path('Base58CheckDecode.py')
        engine = TestEngine()
        arg = base58.b58encode_check('unit test'.encode('utf-8'))
        result = self.run_smart_contract(engine, path, 'main', arg)
        self.assertEqual('unit test', result)

        arg = base58.b58encode_check(''.encode('utf-8'))
        result = self.run_smart_contract(engine, path, 'main', arg)
        self.assertEqual('', result)

        long_string = ('Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam accumsan magna eu massa '
                       'vulputate bibendum. Aliquam commodo euismod tristique. Sed purus erat, pretium ut interdum '
                       'et, aliquet sed mauris. Curabitur vitae turpis euismod, hendrerit mi a, rhoncus justo. Mauris '
                       'sollicitudin, nisl sit amet feugiat pharetra, odio ligula congue tellus, vel pellentesque '
                       'libero leo id dui. Morbi vel risus vehicula, consectetur mauris eget, gravida ligula. '
                       'Maecenas aliquam velit sit amet nisi ultricies, ac sollicitudin nisi mollis. Lorem ipsum '
                       'dolor sit amet, consectetur adipiscing elit. Ut tincidunt, nisi in ullamcorper ornare, '
                       'est enim dictum massa, id aliquet justo magna in purus.')
        arg = base58.b58encode_check(long_string.encode('utf-8'))
        result = self.run_smart_contract(engine, path, 'main', arg)
        self.assertEqual(long_string, result)

    def test_base58_check_decode_mismatched_type(self):
        path = self.get_contract_path('Base58CheckDecodeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_base58_check_encode(self):
        import base58
        path = self.get_contract_path('Base58CheckEncode.py')
        engine = TestEngine()
        expected_result = base58.b58encode_check('unit test'.encode('utf-8'))
        result = self.run_smart_contract(engine, path, 'main', 'unit test',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        expected_result = base58.b58encode_check(''.encode('utf-8'))
        result = self.run_smart_contract(engine, path, 'main', '',
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        long_string = ('Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam accumsan magna eu massa '
                       'vulputate bibendum. Aliquam commodo euismod tristique. Sed purus erat, pretium ut interdum '
                       'et, aliquet sed mauris. Curabitur vitae turpis euismod, hendrerit mi a, rhoncus justo. Mauris '
                       'sollicitudin, nisl sit amet feugiat pharetra, odio ligula congue tellus, vel pellentesque '
                       'libero leo id dui. Morbi vel risus vehicula, consectetur mauris eget, gravida ligula. '
                       'Maecenas aliquam velit sit amet nisi ultricies, ac sollicitudin nisi mollis. Lorem ipsum '
                       'dolor sit amet, consectetur adipiscing elit. Ut tincidunt, nisi in ullamcorper ornare, '
                       'est enim dictum massa, id aliquet justo magna in purus.')
        expected_result = base58.b58encode_check(long_string.encode('utf-8'))
        result = self.run_smart_contract(engine, path, 'main', long_string,
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_base58_check_encode_mismatched_type(self):
        path = self.get_contract_path('Base58CheckEncodeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_serialize_int(self):
        path = self.get_contract_path('SerializeInt.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'serialize_int',
                                         expected_result_type=bytes)
        expected_result = serialize(42)
        self.assertEqual(expected_result, result)

    def test_serialize_bool(self):
        path = self.get_contract_path('SerializeBool.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'serialize_bool',
                                         expected_result_type=bytes)
        expected_result = serialize(True)
        self.assertEqual(expected_result, result)

    def test_serialize_str(self):
        path = self.get_contract_path('SerializeStr.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'serialize_str',
                                         expected_result_type=bytes)
        expected_result = serialize('42')
        self.assertEqual(expected_result, result)

    def test_serialize_sequence(self):
        path = self.get_contract_path('SerializeSequence.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'serialize_sequence',
                                         expected_result_type=bytes)
        expected_result = serialize([2, 3, 5, 7])
        self.assertEqual(expected_result, result)

    def test_serialize_dict(self):
        path = self.get_contract_path('SerializeDict.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'serialize_dict',
                                         expected_result_type=bytes)
        expected_result = serialize({1: 1, 2: 1, 3: 2})
        self.assertEqual(expected_result, result)

    def test_deserialize(self):
        path = self.get_contract_path('Deserialize.py')
        engine = TestEngine()

        expected_result = 42
        value = serialize(expected_result)
        result = self.run_smart_contract(engine, path, 'deserialize_arg', value,
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        expected_result = True
        value = serialize(expected_result)
        result = self.run_smart_contract(engine, path, 'deserialize_arg', value)
        self.assertEqual(expected_result, result)

        value = StackItemType.Boolean + value[1:]
        result = self.run_smart_contract(engine, path, 'deserialize_arg', value)
        self.assertEqual(expected_result, result)

        expected_result = '42'
        value = serialize(expected_result)
        result = self.run_smart_contract(engine, path, 'deserialize_arg', value)
        self.assertEqual(expected_result, result)

        expected_result = b'42'
        value = serialize(expected_result)
        result = self.run_smart_contract(engine, path, 'deserialize_arg', value,
                                         expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        expected_result = [1, '2', b'3']
        value = serialize(expected_result)
        result = self.run_smart_contract(engine, path, 'deserialize_arg', value)
        expected_result[2] = String.from_bytes(expected_result[2])
        self.assertEqual(expected_result, result)

        expected_result = {'int': 1, 'str': '2', 'bytes': b'3'}
        value = serialize(expected_result)
        result = self.run_smart_contract(engine, path, 'deserialize_arg', value)
        expected_result['bytes'] = String.from_bytes(expected_result['bytes'])
        self.assertEqual(expected_result, result)

    def test_deserialize_mismatched_type(self):
        path = self.get_contract_path('DeserializeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_json_serialize(self):
        path = self.get_contract_path('JsonSerialize.py')

        engine = TestEngine()
        test_input = {"one": 1, "two": 2, "three": 3}
        expected_result = json.dumps(test_input, separators=(',', ':'))
        result = self.run_smart_contract(engine, path, 'main', test_input)
        self.assertEqual(expected_result, result)

    def test_json_serialize_int(self):
        path = self.get_contract_path('JsonSerializeInt.py')

        engine = TestEngine()
        expected_result = json.dumps(10)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_json_serialize_bool(self):
        path = self.get_contract_path('JsonSerializeBool.py')

        engine = TestEngine()
        expected_result = json.dumps(True)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_json_serialize_str(self):
        path = self.get_contract_path('JsonSerializeStr.py')

        engine = TestEngine()
        expected_result = json.dumps('unit test')
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_json_serialize_bytes(self):
        path = self.get_contract_path('JsonSerializeBytes.py')

        engine = TestEngine()
        # Python does not accept bytes as parameter for json.dumps() method, since string and bytes ends up being the
        # same on Neo, it's being converted to string, before using dumps
        expected_result = json.dumps(String().from_bytes(b'unit test'))
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_json_deserialize(self):
        path = self.get_contract_path('JsonDeserialize.py')
        engine = TestEngine()

        test_input = json.dumps(12345)
        expected_result = json.loads(test_input)
        result = self.run_smart_contract(engine, path, 'main', test_input)
        self.assertEqual(expected_result, result)

        test_input = json.dumps('unit test')
        expected_result = json.loads(test_input)
        result = self.run_smart_contract(engine, path, 'main', test_input)
        self.assertEqual(expected_result, result)

        test_input = json.dumps(True)
        expected_result = json.loads(test_input)
        result = self.run_smart_contract(engine, path, 'main', test_input)
        self.assertEqual(expected_result, result)

    def test_boa2_serialization_test1(self):
        path = self.get_contract_path('SerializationBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, expected_result_type=bytes)
        expected_result = serialize(['a', 3, ['j', 3, 5], 'jk', 'lmnopqr'])
        self.assertEqual(expected_result, result)

    def test_boa2_serialization_test2(self):
        path = self.get_contract_path('SerializationBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 2, expected_result_type=bytes)
        expected_result = serialize(['a', 3, ['j', 3, 5], 'jk', 'lmnopqr'])
        self.assertEqual(expected_result, result)

    def test_boa2_serialization_test3(self):
        path = self.get_contract_path('SerializationBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertEqual(['a', 3, ['j', 3, 5], 'jk', 'lmnopqr'], result)

    def test_boa2_serialization_test4(self):
        path = self.get_contract_path('SerializationBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 4)
        self.assertEqual(['j', 3, 5], result)

    def test_atoi(self):
        path = self.get_contract_path('Atoi.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', '10', 10)
        self.assertEqual(10, result)

        result = self.run_smart_contract(engine, path, 'main', '10', 16)
        self.assertEqual(16, result)

        result = self.run_smart_contract(engine, path, 'main', '123', 10)
        self.assertEqual(123, result)

        result = self.run_smart_contract(engine, path, 'main', '123', 16)
        self.assertEqual(291, result)

        result = self.run_smart_contract(engine, path, 'main', '1f', 16)
        self.assertEqual(31, result)

        result = self.run_smart_contract(engine, path, 'main', 'ff', 16)
        self.assertEqual(-1, result)

        with self.assertRaisesRegex(TestExecutionException, self.CANT_PARSE_VALUE_MSG):
            self.run_smart_contract(engine, path, 'main', 'string', 10)

        with self.assertRaisesRegex(TestExecutionException, self.CANT_PARSE_VALUE_MSG):
            self.run_smart_contract(engine, path, 'main', 'string', 16)

        with self.assertRaisesRegex(TestExecutionException, self.CANT_PARSE_VALUE_MSG):
            self.run_smart_contract(engine, path, 'main', 'abc', 10)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.ARGUMENT_OUT_OF_RANGE_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'main', '10', 2)

    def test_atoi_default(self):
        path = self.get_contract_path('AtoiDefault.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', '10')
        self.assertEqual(10, result)

        result = self.run_smart_contract(engine, path, 'main', '123')
        self.assertEqual(123, result)

        with self.assertRaisesRegex(TestExecutionException, self.CANT_PARSE_VALUE_MSG):
            self.run_smart_contract(engine, path, 'main', 'string')

    def test_atoi_too_few_parameters(self):
        path = self.get_contract_path('AtoiTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_atoi_too_many_parameters(self):
        path = self.get_contract_path('AtoiTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_atoi_mismatched_type(self):
        path = self.get_contract_path('AtoiMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_itoa(self):
        path = self.get_contract_path('Itoa')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 10, 10)
        self.assertEqual('10', result)
        result = self.run_smart_contract(engine, path, 'main', 16, 16)
        self.assertEqual('10', result)
        result = self.run_smart_contract(engine, path, 'main', -1, 10)
        self.assertEqual('-1', result)
        result = self.run_smart_contract(engine, path, 'main', -1, 16)
        self.assertEqual('f', result)
        result = self.run_smart_contract(engine, path, 'main', 15, 16)
        self.assertEqual('0f', result)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.ARGUMENT_OUT_OF_RANGE_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'main', 10, 2)

    def test_itoa_default(self):
        path = self.get_contract_path('ItoaDefault')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 10)
        self.assertEqual('10', result)
        result = self.run_smart_contract(engine, path, 'main', -1)
        self.assertEqual('-1', result)

    def test_itoa_too_few_arguments(self):
        path = self.get_contract_path('ItoaTooFewArguments')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_itoa_too_many_arguments(self):
        path = self.get_contract_path('ItoaTooManyArguments')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_itoa_mismatched_type(self):
        path = self.get_contract_path('ItoaMismatchedType')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_memory_search(self):
        path = self.get_contract_path('MemorySearch')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'abcde', 'a', 0, False)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 0, False)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'b', 0, False)
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'c', 0, False)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'd', 0, False)
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'e', 0, False)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 1, False)
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'cd', 0, False)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'abe', 0, False)
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'main', b'aaaaa', b'a', 0, False)
        self.assertEqual(0, result)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.ARGUMENT_OUT_OF_RANGE_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 20, False)

    def test_memory_search_backward(self):
        path = self.get_contract_path('MemorySearch')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'abcde', 'a', 5, True)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 5, True)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'b', 5, True)
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'c', 5, True)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'd', 5, True)
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'e', 5, True)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 0, True)
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'cd', 5, True)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'abe', 5, True)
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'main', b'aaaaa', b'a', 5, True)
        self.assertEqual(4, result)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.ARGUMENT_OUT_OF_RANGE_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 20, True)

    def test_memory_search_start(self):
        path = self.get_contract_path('MemorySearchStart')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'abcde', 'a', 0)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 0)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'e', 0)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 1)
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'cd', 0)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'abe', 0)
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'main', b'aaaaa', b'a', 0)
        self.assertEqual(0, result)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.ARGUMENT_OUT_OF_RANGE_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'main', b'abcde', b'a', 20)

    def test_memory_search_default_values(self):
        path = self.get_contract_path('MemorySearchDefault')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'abcde', 'a')
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'a')
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'b')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'c')
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'd')
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'e')
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'cd')
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', b'abcde', b'aa')
        self.assertEqual(-1, result)

    def test_memory_search_mismatched_type(self):
        path = self.get_contract_path('MemorySearchMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_memory_search_too_few_parameters(self):
        path = self.get_contract_path('MemorySearchTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_memory_search_too_many_parameters(self):
        path = self.get_contract_path('MemorySearchTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_memory_compare(self):
        path = self.get_contract_path('MemoryCompare')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'abc', 'abc')
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', 'abc', 'ABC')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'main', 'ABC', 'abc')
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'main', b'abc', b'abc')
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', b'abc', b'ABC')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'main', b'ABC', b'abc')
        self.assertEqual(-1, result)

    def test_memory_compare_too_few_parameters(self):
        path = self.get_contract_path('MemoryCompareTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_memory_compare_too_many_parameters(self):
        path = self.get_contract_path('MemoryCompareTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_memory_compare_mismatched_type(self):
        path = self.get_contract_path('MemoryCompareMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)



from typing import List, Tuple

from boa3.model.operation.operator import Operator
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.type.type import IType, Type
from boa3.neo.vm.opcode.Opcode import Opcode


class Positive(UnaryOperation):
    """
    A class used to represent a numeric positive operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar operand: the operand type. Inherited from :class:`UnaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.int]

    def __init__(self, operand: IType = Type.int):
        self.operator: Operator = Operator.Plus
        super().__init__(operand)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        operand: IType = types[0]

        return any(_type.is_type_of(operand) for _type in self._valid_types)

    def _get_result(self, operand: IType) -> IType:
        if self.validate_type(operand):
            return operand
        else:
            return Type.none

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        # it is the identity function, so there is no need of including another opcode
        return []



import boa3_test.test_sc.import_test.variable_import.GenerateImportedVariable as OtherModule
from boa3.builtin.compile_time import public


@public
def get_foo() -> bytes:
    return OtherModule.foo()


@public
def get_bar() -> str:
    return OtherModule.bar()



from typing import List, Optional, Tuple

from boa3.model.builtin.classmethod.countsequencemethod import CountSequenceMethod
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class CountSequenceGenericMethod(CountSequenceMethod):

    def __init__(self, sequence_type: Optional[SequenceType] = None, arg_value: Optional[IType] = None):
        super().__init__(sequence_type, arg_value)

    def generic_verification(self, inc_statement_bytes=None,
                             get_equals_statement_bytes=None) -> List[Tuple[Opcode, bytes]]:

        if self._generic_verification_opcodes is None:
            jmp_place_holder = (Opcode.JMP, b'\x01')

            from boa3.compiler.codegenerator import get_bytes_count
            from boa3.model.type.type import Type

            sequence_verify_value_is_sequence = [  # verifies if value is a Sequence
                (Opcode.PUSH3, b''),
                (Opcode.PICK, b''),
                (Opcode.ISTYPE, Type.sequence.stack_item),
                jmp_place_holder,  # JMP to sequence_get_element if not
            ]

            sequence_verify_item_is_sequence = [  # verifies if the sequence[index] is a Sequence
                (Opcode.PUSH2, b''),
                (Opcode.PICK, b''),
                (Opcode.OVER, b''),
                (Opcode.PICKITEM, b''),
                (Opcode.ISTYPE, Type.sequence.stack_item),
                jmp_place_holder,  # jmp to list_tuple_count_index_dec if not
            ]

            sequence_verify_item_value_is_same_size = [  # verify if value and sequence[index] are the same size
                (Opcode.PUSH3, b''),
                (Opcode.PICK, b''),
                (Opcode.PUSH3, b''),
                (Opcode.PICK, b''),
                (Opcode.PUSH2, b''),
                (Opcode.PICK, b''),
                (Opcode.PICKITEM, b''),
                (Opcode.OVER, b''),
                (Opcode.SIZE, b''),
                (Opcode.OVER, b''),
                (Opcode.SIZE, b''),
                jmp_place_holder  # jmp to sequence_compare_sequences_initialize if true
            ]

            sequence_remove_aux_and_verify_next = [  # remove auxiliary values on stack and verify next item
                (Opcode.DROP, b''),
                (Opcode.DROP, b''),
                jmp_place_holder  # jumps to list_tuple_count_index_dec
            ]

            sequence_compare_sequences_initialize = [
                # starts another index for both sequences (value and sequence[índex])
                (Opcode.DUP, b''),
                (Opcode.SIZE, b''),
            ]

            sequence_compare_sequences_start = [  # compare both sequences
                (Opcode.DEC, b''),
                (Opcode.OVER, b''),
                (Opcode.OVER, b''),
                (Opcode.PICKITEM, b''),
                (Opcode.PUSH3, b''),
                (Opcode.PICK, b''),
                (Opcode.PUSH2, b''),
                (Opcode.PICK, b''),
                (Opcode.PICKITEM, b''),
                jmp_place_holder  # jumps to #sequence_compare_sequences_end_is_not_equal if they have different items
            ]

            sequence_compare_sequences_verify_end = [  # verify if all items were compared already
                (Opcode.DUP, b''),
                (Opcode.PUSH0, b''),
                # jumps to back to the verification if not
            ]

            sequence_compare_sequences_end_is_equal = [  # if they have the same values, then increment count
                (Opcode.DROP, b''),  # remove auxiliary values on stack
                (Opcode.DROP, b''),
                (Opcode.DROP, b''),
                jmp_place_holder  # jumps to sequence_count_inc
            ]

            sequence_compare_sequences_end_is_not_equal = [
                # if they do not have the same values go check the next item
                (Opcode.DROP, b''),
                (Opcode.DROP, b''),
                (Opcode.DROP, b''),
                jmp_place_holder  # jumps to list_tuple_count_index_dec
            ]

            num_jmp_code = inc_statement_bytes + get_equals_statement_bytes
            sequence_compare_sequences_end_is_not_equal[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code, True)

            num_jmp_code = get_bytes_count(sequence_compare_sequences_end_is_not_equal) + get_equals_statement_bytes
            sequence_compare_sequences_end_is_equal[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code, True)

            num_jmp_code = -get_bytes_count(sequence_compare_sequences_verify_end + sequence_compare_sequences_start)
            sequence_compare_sequences_verify_end.append(OpcodeHelper.get_jump_and_data(Opcode.JMPGT, num_jmp_code))

            num_jmp_code = get_bytes_count(sequence_compare_sequences_verify_end +
                                           sequence_compare_sequences_end_is_equal)
            sequence_compare_sequences_start[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPNE, num_jmp_code, True)

            num_jmp_code = get_bytes_count(
                sequence_compare_sequences_initialize + sequence_compare_sequences_start +
                sequence_compare_sequences_verify_end + sequence_compare_sequences_end_is_equal +
                sequence_compare_sequences_end_is_not_equal) + get_equals_statement_bytes + inc_statement_bytes
            sequence_remove_aux_and_verify_next[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code, True)

            num_jmp_code = get_bytes_count(sequence_remove_aux_and_verify_next)
            sequence_verify_item_value_is_same_size[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPEQ, num_jmp_code, True)

            num_jmp_code = get_bytes_count(
                sequence_verify_item_value_is_same_size + sequence_remove_aux_and_verify_next +
                sequence_compare_sequences_initialize + sequence_compare_sequences_start +
                sequence_compare_sequences_verify_end + sequence_compare_sequences_end_is_equal +
                sequence_compare_sequences_end_is_not_equal) + get_equals_statement_bytes + inc_statement_bytes
            sequence_verify_item_is_sequence[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)

            num_jmp_code = get_bytes_count(sequence_verify_item_is_sequence + sequence_verify_item_value_is_same_size +
                                           sequence_remove_aux_and_verify_next + sequence_compare_sequences_initialize +
                                           sequence_compare_sequences_start + sequence_compare_sequences_verify_end +
                                           sequence_compare_sequences_end_is_equal +
                                           sequence_compare_sequences_end_is_not_equal)
            sequence_verify_value_is_sequence[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)

            self._generic_verification_opcodes = (
                sequence_verify_value_is_sequence +
                sequence_verify_item_is_sequence +
                sequence_verify_item_value_is_same_size +
                sequence_remove_aux_and_verify_next +
                sequence_compare_sequences_initialize +
                sequence_compare_sequences_start +
                sequence_compare_sequences_verify_end +
                sequence_compare_sequences_end_is_equal +
                sequence_compare_sequences_end_is_not_equal
            )

        return super().generic_verification()



import ast
from typing import Any, Dict, List, Optional

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable


class IntMethod(IBuiltinMethod):

    def __init__(self, args: Dict[str, Variable] = None, defaults: List[ast.AST] = None):
        from boa3.model.type.type import Type
        identifier = 'int'
        super().__init__(identifier, args, defaults=defaults, return_type=Type.int)

    @property
    def _arg_value(self) -> Variable:
        return self.args['value']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type

        if self._arg_value.type is Type.int:
            return '-{0}_{1}'.format(self._identifier, Type.str.identifier + Type.bytes.identifier)

        if self._arg_value.type is Type.str or self._arg_value.type is Type.bytes:
            return '-{0}_{1}'.format(self._identifier, Type.sequence.identifier)

        return self._identifier

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, list):
            value = [value]

        from boa3.model.type.type import Type
        from boa3.model.builtin.builtin import Builtin
        if len(value) == 0 or Type.int.is_type_of(value[0]):
            return Builtin.IntInt

        if Type.str.is_type_of(value[0]) or Type.bytes.is_type_of(value[0]):
            return Builtin.IntByteString



from typing import Dict

from boa3.model.builtin.interop.nativecontract import LedgerMethod
from boa3.model.variable import Variable


class GetTransactionHeightMethod(LedgerMethod):

    def __init__(self):
        from boa3.model.type.collection.sequence.uint256type import UInt256Type
        from boa3.model.type.type import Type

        identifier = 'get_transaction_height'
        syscall = 'getTransactionHeight'
        args: Dict[str, Variable] = {'hash_': Variable(UInt256Type.build())}
        super().__init__(identifier, syscall, args, return_type=Type.int)



from boa3.builtin.compile_time import public

from boa3_test.test_sc.import_test.FromImportTyping import *


@public
def call_imported_method() -> list:
    a: List[Any] = EmptyList()
    return a


@public
def call_imported_variable() -> list:
    return empty_list



__all__ = ['CollectionMembership',
           'CollectionNotMembership'
           ]

from boa3.model.operation.binary.additional.membership import CollectionMembership
from boa3.model.operation.binary.additional.notmembership import CollectionNotMembership



from abc import ABC

from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.method import Method
from boa3.model.property import Property


class IBuiltinProperty(Property, IdentifiedSymbol, ABC):
    def __init__(self, identifier: str, getter: Method, setter: Method = None):
        super().__init__(getter, setter)
        self._identifier = identifier



from typing import List, Tuple

from boa3.model.builtin.method.printmethod import PrintMethod
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class PrintBoolMethod(PrintMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        arg_value = Type.bool

        super().__init__(arg_value)

    @property
    def print_value_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        if self._print_value_opcodes is None:
            from boa3.compiler.codegenerator import get_bytes_count
            if_value_is_true = [
                OpcodeHelper.get_pushdata_and_data('True')
            ]
            if_value_is_false = [
                OpcodeHelper.get_pushdata_and_data('False'),
                OpcodeHelper.get_jump_and_data(Opcode.JMP, get_bytes_count(if_value_is_true), jump_through=True)
            ]

            self._print_value_opcodes = (
                [
                    (Opcode.NZ, b''),
                    OpcodeHelper.get_jump_and_data(Opcode.JMPIF, get_bytes_count(if_value_is_false), jump_through=True),
                ] + if_value_is_false
                + if_value_is_true
            )

        return super().print_value_opcodes



import ast
from typing import Dict, List

from boa3.model import set_internal_call
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class LoadScriptMethod(InteropMethod):

    def __init__(self):
        from boa3.model.builtin.interop.contract.callflagstype import CallFlagsType
        from boa3.model.type.type import Type
        from boa3.model.type.primitive.bytestringtype import ByteStringType
        from boa3.neo3.contracts import CallFlags

        identifier = 'load_script'
        syscall = 'System.Runtime.LoadScript'
        call_flags: CallFlagsType = CallFlagsType.build()

        args: Dict[str, Variable] = {
            'script': Variable(ByteStringType.build()),
            'args': Variable(Type.sequence),
            'call_flags': Variable(call_flags)
        }

        args_default = ast.parse("{0}".format(Type.sequence.default_value)
                                 ).body[0].value
        call_flags_default = set_internal_call(ast.parse("{0}.{1}".format(call_flags.identifier,
                                                                          CallFlags.NONE.name)
                                                         ).body[0].value)

        super().__init__(identifier, syscall, args, defaults=[args_default, call_flags_default], return_type=Type.any)

    @property
    def generation_order(self) -> List[int]:
        indexes = super().generation_order
        context_index = list(self.args).index('args')

        if indexes[-1] != context_index:
            # args must be the first generated argument
            indexes.remove(context_index)
            indexes.insert(0, context_index)

        return indexes



import json

from boa3 import constants
from boa3.boa3 import Boa3
from boa3.exception import CompilerError, CompilerWarning
from boa3.model.builtin.interop.interop import Interop
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.contract.neomanifeststruct import NeoManifestStruct
from boa3_test.tests.test_classes.testengine import TestEngine


class TestContractInterop(BoaTest):
    default_folder: str = 'test_sc/interop_test/contract'

    def test_call_contract(self):
        path = self.get_contract_path('CallScriptHash.py')
        call_contract_path = self.get_contract_path('test_sc/arithmetic_test', 'Addition.py')

        engine = TestEngine()
        expected_output = self.run_smart_contract(engine, call_contract_path, 'add', 1, 2)
        call_hash = engine.executed_script_hash.to_array()

        engine.reset_engine()
        call_contract_path = call_contract_path.replace('.py', '.nef')
        with self.assertRaisesRegex(TestExecutionException, self.CALLED_CONTRACT_DOES_NOT_EXIST_MSG):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'add', [1, 2])
        engine.add_contract(call_contract_path)

        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'add', [1, 2])
        self.assertEqual(expected_output, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'add', [-42, -24])
        self.assertEqual(-66, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'add', [-42, 24])
        self.assertEqual(-18, result)

    def test_call_contract_with_cast(self):
        path = self.get_contract_path('CallScriptHashWithCast.py')
        call_contract_path = self.get_contract_path('test_sc/arithmetic_test', 'Addition.py')

        engine = TestEngine()
        self.run_smart_contract(engine, call_contract_path, 'add', 1, 2)
        call_hash = engine.executed_script_hash.to_array()

        engine.reset_engine()
        call_contract_path = call_contract_path.replace('.py', '.nef')
        with self.assertRaisesRegex(TestExecutionException, self.CALLED_CONTRACT_DOES_NOT_EXIST_MSG):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'add', [1, 2])
        engine.add_contract(call_contract_path)

        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'add', [1, 2])
        self.assertEqual(True, result)

    def test_call_contract_without_args(self):
        path = self.get_contract_path('CallScriptHashWithoutArgs.py')
        call_contract_path = self.get_contract_path('test_sc/list_test', 'IntList.py')
        Boa3.compile_and_save(call_contract_path)

        engine = TestEngine()
        expected_output = self.run_smart_contract(engine, call_contract_path, 'Main')
        call_hash = engine.executed_script_hash.to_array()

        engine.reset_engine()
        call_contract_path = call_contract_path.replace('.py', '.nef')
        with self.assertRaisesRegex(TestExecutionException, self.CALLED_CONTRACT_DOES_NOT_EXIST_MSG):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'Main')
        engine.add_contract(call_contract_path)

        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'Main')
        self.assertEqual(expected_output, result)

    def test_call_contract_with_flags(self):
        path = self.get_contract_path('CallScriptHashWithFlags.py')
        call_contract_path = self.get_contract_path('CallFlagsUsage.py')

        engine = TestEngine()
        self.run_smart_contract(engine, call_contract_path, 'notify_user')
        call_hash = engine.executed_script_hash.to_array()

        engine.reset_engine()
        call_contract_path = call_contract_path.replace('.py', '.nef')
        with self.assertRaisesRegex(TestExecutionException, f'^{self.ARGUMENT_OUT_OF_RANGE_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'Main')
        engine.add_contract(call_contract_path)

        from boa3.neo3.contracts import CallFlags

        with self.assertRaisesRegex(TestExecutionException, f'^{self.CANT_CALL_SYSCALL_WITH_FLAG_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'get_value', ['num'], CallFlags.NONE)

        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'get_value', ['num'], CallFlags.READ_ONLY)
        self.assertEqual(0, result)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.CANT_CALL_SYSCALL_WITH_FLAG_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'put_value', ['num', 10], CallFlags.READ_ONLY)
            self.run_smart_contract(engine, path, 'Main', call_hash, 'put_value', ['num', 10], CallFlags.NONE)

        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'put_value', ['num', 10], CallFlags.STATES)
        self.assertEqual(None, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'get_value', ['num'], CallFlags.READ_ONLY)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'put_value', ['num', 99], CallFlags.ALL)
        self.assertEqual(None, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'get_value', ['num'], CallFlags.READ_ONLY)
        self.assertEqual(99, result)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.CANT_CALL_SYSCALL_WITH_FLAG_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'notify_user', [], CallFlags.READ_ONLY)
            self.run_smart_contract(engine, path, 'Main', call_hash, 'notify_user', [], CallFlags.STATES)
            self.run_smart_contract(engine, path, 'Main', call_hash, 'notify_user', [], CallFlags.NONE)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'notify_user', [], CallFlags.ALL)
        self.assertEqual(None, result)
        notify = engine.get_events(origin=call_hash)
        self.assertEqual(1, len(notify))
        self.assertEqual('Notify was called', notify[0].arguments[0])

        with self.assertRaisesRegex(TestExecutionException, f'^{self.CANT_CALL_SYSCALL_WITH_FLAG_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'call_another_contract', [], CallFlags.STATES)
            self.run_smart_contract(engine, path, 'Main', call_hash, 'call_another_contract', [], CallFlags.NONE)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'call_another_contract', [], CallFlags.ALL)
        self.assertEqual(0, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'call_another_contract',
                                         [], CallFlags.READ_ONLY)
        self.assertEqual(0, result)

    def test_call_contract_too_many_parameters(self):
        path = self.get_contract_path('CallScriptHashTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_call_contract_too_few_parameters(self):
        path = self.get_contract_path('CallScriptHashTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_create_contract(self):
        path = self.get_contract_path('CreateContract.py')
        call_contract_path = self.get_contract_path('test_sc/arithmetic_test', 'Addition.py')
        Boa3.compile_and_save(call_contract_path)

        nef_file, manifest = self.get_bytes_output(call_contract_path)
        arg_manifest = String(json.dumps(manifest, separators=(',', ':'))).to_bytes()

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', nef_file, arg_manifest, None)

        self.assertEqual(5, len(result))
        self.assertEqual(nef_file, result[3])
        manifest_struct = NeoManifestStruct.from_json(manifest)
        self.assertEqual(manifest_struct, result[4])

    def test_create_contract_data_deploy(self):
        path = self.get_contract_path('CreateContract.py')
        call_contract_path = self.get_contract_path('NewContract.py')
        Boa3.compile_and_save(call_contract_path)

        nef_file, manifest = self.get_bytes_output(call_contract_path)
        arg_manifest = String(json.dumps(manifest, separators=(',', ':'))).to_bytes()

        engine = TestEngine()
        data = 'some sort of data'
        result = self.run_smart_contract(engine, path, 'Main', nef_file, arg_manifest, data)

        self.assertEqual(5, len(result))
        self.assertEqual(nef_file, result[3])
        manifest_struct = NeoManifestStruct.from_json(manifest)
        self.assertEqual(manifest_struct, result[4])

        notifies = engine.get_events('notify')
        self.assertEqual(2, len(notifies))
        self.assertEqual(False, notifies[0].arguments[0])  # not updated
        self.assertEqual(data, notifies[1].arguments[0])  # data
        result = self.run_smart_contract(engine, call_contract_path, 'main')
        self.assertEqual(data, result)

    def test_create_contract_too_many_parameters(self):
        path = self.get_contract_path('CreateContractTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_create_contract_too_few_parameters(self):
        path = self.get_contract_path('CreateContractTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_update_contract(self):
        path = self.get_contract_path('UpdateContract.py')
        engine = TestEngine()
        with self.assertRaisesRegex(TestExecutionException, f'{self.CANT_FIND_METHOD_MSG_PREFIX} : new_method'):
            self.run_smart_contract(engine, path, 'new_method')

        new_path = self.get_contract_path('test_sc/interop_test', 'UpdateContract.py')
        self.compile_and_save(new_path)
        new_nef, new_manifest = self.get_bytes_output(new_path)
        arg_manifest = String(json.dumps(new_manifest, separators=(',', ':'))).to_bytes()

        result = self.run_smart_contract(engine, path, 'update', new_nef, arg_manifest, None)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'new_method')
        self.assertEqual(42, result)

    def test_update_contract_data_deploy(self):
        path = self.get_contract_path('UpdateContract.py')
        engine = TestEngine()
        with self.assertRaisesRegex(TestExecutionException, f'{self.CANT_FIND_METHOD_MSG_PREFIX} : new_method'):
            self.run_smart_contract(engine, path, 'new_method')

        new_path = self.get_contract_path('test_sc/interop_test', 'UpdateContract.py')
        self.compile_and_save(new_path)
        new_nef, new_manifest = self.get_bytes_output(new_path)
        arg_manifest = String(json.dumps(new_manifest, separators=(',', ':'))).to_bytes()

        data = 'this function was deployed'
        result = self.run_smart_contract(engine, path, 'update', new_nef, arg_manifest, data)
        self.assertIsVoid(result)
        notifies = engine.get_events('notify')
        self.assertEqual(2, len(notifies))
        self.assertEqual(True, notifies[0].arguments[0])
        self.assertEqual(data, notifies[1].arguments[0])

    def test_update_contract_too_many_parameters(self):
        path = self.get_contract_path('UpdateContractTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_update_contract_too_few_parameters(self):
        path = self.get_contract_path('UpdateContractTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_destroy_contract(self):
        path = self.get_contract_path('DestroyContract.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

        script_hash = engine.executed_script_hash.to_array()
        call_contract_path = self.get_contract_path('CallScriptHash.py')
        with self.assertRaisesRegex(TestExecutionException, f'^{self.CALLED_CONTRACT_DOES_NOT_EXIST_MSG}'):
            self.run_smart_contract(engine, call_contract_path, 'Main',
                                    script_hash, 'Main', [])

    def test_destroy_contract_too_many_parameters(self):
        path = self.get_contract_path('DestroyContractTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_get_neo_native_script_hash(self):
        value = constants.NEO_SCRIPT
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array()
            + value
            + Opcode.RET
        )

        path = self.get_contract_path('NeoScriptHash.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(value, result)

    def test_neo_native_script_hash_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('NeoScriptHashCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_get_gas_native_script_hash(self):
        value = constants.GAS_SCRIPT
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array()
            + value
            + Opcode.RET
        )

        path = self.get_contract_path('GasScriptHash.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(value, result)

    def test_gas_native_script_hash_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('GasScriptHashCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_call_flags_type(self):
        path = self.get_contract_path('CallFlagsType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'ALL')
        self.assertEqual(0b00001111, result)
        result = self.run_smart_contract(engine, path, 'main', 'READ_ONLY')
        self.assertEqual(0b00000101, result)
        result = self.run_smart_contract(engine, path, 'main', 'STATES')
        self.assertEqual(0b00000011, result)
        result = self.run_smart_contract(engine, path, 'main', 'ALLOW_NOTIFY')
        self.assertEqual(0b00001000, result)
        result = self.run_smart_contract(engine, path, 'main', 'ALLOW_CALL')
        self.assertEqual(0b00000100, result)
        result = self.run_smart_contract(engine, path, 'main', 'WRITE_STATES')
        self.assertEqual(0b00000010, result)
        result = self.run_smart_contract(engine, path, 'main', 'READ_STATES')
        self.assertEqual(0b00000001, result)
        result = self.run_smart_contract(engine, path, 'main', 'NONE')
        self.assertEqual(0, result)

    def test_get_call_flags(self):
        path = self.get_contract_path('CallScriptHashWithFlags.py')
        call_contract_path = self.get_contract_path('GetCallFlags.py')
        Boa3.compile_and_save(call_contract_path)

        engine = TestEngine()
        self.run_smart_contract(engine, call_contract_path, 'main')
        call_hash = engine.executed_script_hash.to_array()

        engine.reset_engine()
        call_contract_path = call_contract_path.replace('.py', '.nef')
        with self.assertRaisesRegex(TestExecutionException, f'^{self.ARGUMENT_OUT_OF_RANGE_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'Main', call_hash, 'main')
        engine.add_contract(call_contract_path)

        from boa3.neo3.contracts import CallFlags

        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'main', [], CallFlags.ALL)
        self.assertEqual(CallFlags.ALL, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'main', [], CallFlags.READ_ONLY)
        self.assertEqual(CallFlags.READ_ONLY, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'main', [], CallFlags.STATES)
        self.assertEqual(CallFlags.STATES, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'main', [], CallFlags.NONE)
        self.assertEqual(CallFlags.NONE, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'main', [], CallFlags.READ_STATES)
        self.assertEqual(CallFlags.READ_STATES, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'main', [], CallFlags.WRITE_STATES)
        self.assertEqual(CallFlags.WRITE_STATES, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'main', [], CallFlags.ALLOW_CALL)
        self.assertEqual(CallFlags.ALLOW_CALL, result)
        result = self.run_smart_contract(engine, path, 'Main', call_hash, 'main', [], CallFlags.ALLOW_NOTIFY)
        self.assertEqual(CallFlags.ALLOW_NOTIFY, result)

    def test_import_contract(self):
        path = self.get_contract_path('ImportContract.py')
        call_contract_path = self.get_contract_path('test_sc/arithmetic_test', 'Addition.py')

        engine = TestEngine()
        expected_output = self.run_smart_contract(engine, call_contract_path, 'add', 1, 2)
        call_hash = engine.executed_script_hash.to_array()

        engine.reset_engine()
        call_contract_path = call_contract_path.replace('.py', '.nef')
        engine.add_contract(call_contract_path)

        result = self.run_smart_contract(engine, path, 'main', call_hash, 'add', [1, 2])
        self.assertEqual(expected_output, result)

        result = self.run_smart_contract(engine, path, 'call_flags_all')
        from boa3.neo3.contracts import CallFlags
        self.assertEqual(CallFlags.ALL, result)

    def test_import_interop_contract(self):
        path = self.get_contract_path('ImportInteropContract.py')
        call_contract_path = self.get_contract_path('test_sc/arithmetic_test', 'Addition.py')

        engine = TestEngine()
        expected_output = self.run_smart_contract(engine, call_contract_path, 'add', 1, 2)
        call_hash = engine.executed_script_hash.to_array()

        engine.reset_engine()
        call_contract_path = call_contract_path.replace('.py', '.nef')
        engine.add_contract(call_contract_path)

        result = self.run_smart_contract(engine, path, 'main', call_hash, 'add', [1, 2])
        self.assertEqual(expected_output, result)

        result = self.run_smart_contract(engine, path, 'call_flags_all')
        from boa3.neo3.contracts import CallFlags
        self.assertEqual(CallFlags.ALL, result)

    def test_create_standard_account(self):
        from boa3.neo.vm.type.StackItem import StackItemType
        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x01'
            + Opcode.LDARG0
            + Opcode.CONVERT
            + StackItemType.ByteString
            + Opcode.DUP
            + Opcode.ISNULL
            + Opcode.JMPIF
            + Integer(8).to_byte_array(min_length=1)
            + Opcode.DUP
            + Opcode.SIZE
            + Opcode.PUSHINT8
            + Integer(33).to_byte_array(min_length=1)
            + Opcode.JMPEQ
            + Integer(3).to_byte_array(min_length=1)
            + Opcode.THROW
            + Opcode.SYSCALL
            + Interop.CreateStandardAccount.interop_method_hash
            + Opcode.RET
        )
        path = self.get_contract_path('CreateStandardAccount.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_create_standard_account_too_few_parameters(self):
        path = self.get_contract_path('CreateStandardAccountTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_create_standard_account_too_many_parameters(self):
        path = self.get_contract_path('CreateStandardAccountTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_get_minimum_deployment_fee(self):
        path = self.get_contract_path('GetMinimumDeploymentFee.py')
        engine = TestEngine()

        minimum_cost = 10 * 10 ** 8  # minimum deployment cost is 10 GAS right now
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(minimum_cost, result)

    def test_get_minimum_deployment_fee_too_many_parameters(self):
        path = self.get_contract_path('GetMinimumDeploymentFeeTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_create_multisig_account(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x02'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.CreateMultisigAccount.interop_method_hash
            + Opcode.RET
        )
        path = self.get_contract_path('CreateMultisigAccount.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_create_multisig_account_too_few_parameters(self):
        path = self.get_contract_path('CreateMultisigAccountTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_create_multisig_account_too_many_parameters(self):
        path = self.get_contract_path('CreateMultisigAccountTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)



from __future__ import annotations

from typing import Dict, Optional, Tuple, Union

from boa3.constants import FOUR_BYTES_MAX_VALUE
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String


def get_pushdata_and_data(bytestring: Union[str, bytes]) -> Tuple[Opcode, bytes]:
    """
    Gets the push opcode and data to the respective str ot bytes value

    :param bytestring: value that will be pushed
    :return: the respective opcode and its required data
    :rtype: Tuple[Opcode, bytes]
    """
    if isinstance(bytestring, str):
        bytestring = String(bytestring).to_bytes()

    bytes_size = Integer(len(bytestring)).to_byte_array(min_length=1)
    if len(bytes_size) == 1:
        opcode = Opcode.PUSHDATA1
    elif len(bytes_size) == 2:
        opcode = Opcode.PUSHDATA2
    else:
        if len(bytes_size) > 4:
            bytestring = bytestring[:FOUR_BYTES_MAX_VALUE]
        bytes_size = Integer(len(bytestring)).to_byte_array(min_length=4)
        opcode = Opcode.PUSHDATA4

    data = bytes_size + bytestring
    return opcode, data


def get_pushdata_and_data_from_size(data_size: int) -> Tuple[Opcode, bytes]:
    bytes_size = Integer(data_size).to_byte_array(min_length=1)
    if len(bytes_size) == 1:
        opcode = Opcode.PUSHDATA1
    elif len(bytes_size) == 2:
        opcode = Opcode.PUSHDATA2
    else:
        bytes_size = FOUR_BYTES_MAX_VALUE
        opcode = Opcode.PUSHDATA4

    return opcode, bytes_size


def get_literal_push(integer: int) -> Optional[Opcode]:
    """
    Gets the push opcode to the respective integer

    :param integer: value that will be pushed
    :return: the respective opcode
    :rtype: Opcode or None
    """
    if -1 <= integer <= 16:
        opcode_value: int = Integer.from_bytes(Opcode.PUSH0) + integer
        return Opcode(Integer(opcode_value).to_byte_array())
    else:
        return None


def get_push_and_data(integer: int) -> Tuple[Opcode, bytes]:
    """
    Gets the push opcode and data to the respective integer

    :param integer: value that will be pushed
    :return: the respective opcode and its required data
    :rtype: Tuple[Opcode, bytes]
    """
    opcode = get_literal_push(integer)
    if isinstance(opcode, Opcode):
        return opcode, b''
    else:
        data = Integer(integer).to_byte_array(signed=True, min_length=1)
        if len(data) == 1:
            opcode = Opcode.PUSHINT8
        elif len(data) == 2:
            opcode = Opcode.PUSHINT16
        elif len(data) <= 4:
            data = Integer(integer).to_byte_array(signed=True, min_length=4)
            opcode = Opcode.PUSHINT32
        elif len(data) <= 8:
            data = Integer(integer).to_byte_array(signed=True, min_length=8)
            opcode = Opcode.PUSHINT64
        elif len(data) <= 16:
            data = Integer(integer).to_byte_array(signed=True, min_length=16)
            opcode = Opcode.PUSHINT128
        else:
            data = data[:32]
            opcode = Opcode.PUSHINT256

        return opcode, data


def has_larger_opcode(opcode: Opcode) -> bool:
    return opcode in _larger_opcode


def get_larger_opcode(opcode: Opcode):
    """
    Gets the large opcode to the standard opcode

    :return: the respective opcode
    :rtype: Opcode or None
    """
    if opcode in _larger_opcode:
        return _larger_opcode[opcode]
    elif opcode in _larger_opcode.values():
        return opcode
    else:
        return None


_larger_opcode: Dict[Opcode, Opcode] = {
    Opcode.JMP: Opcode.JMP_L,
    Opcode.JMPIF: Opcode.JMPIF_L,
    Opcode.JMPIFNOT: Opcode.JMPIFNOT_L,
    Opcode.JMPEQ: Opcode.JMPEQ_L,
    Opcode.JMPNE: Opcode.JMPNE_L,
    Opcode.JMPGT: Opcode.JMPGT_L,
    Opcode.JMPGE: Opcode.JMPGE_L,
    Opcode.JMPLT: Opcode.JMPLT_L,
    Opcode.JMPLE: Opcode.JMPLE_L,
    Opcode.CALL: Opcode.CALL_L,
    Opcode.TRY: Opcode.TRY_L,
    Opcode.ENDTRY: Opcode.ENDTRY_L
}


def is_jump(opcode: Opcode) -> bool:
    return Opcode.JMP <= opcode <= Opcode.JMPLE_L


def get_jump_and_data(opcode: Opcode, integer: int, jump_through: bool = False) -> Tuple[Opcode, bytes]:
    """
    Gets the jump opcode and data to the respective integer

    :param opcode: which jump will be used
    :param integer: number of bytes that'll be jumped
    :param jump_through: whether it should go over the instructions or not
    :return: the respective opcode and its required data
    :rtype: Tuple[Opcode, bytes]
    """
    if not is_jump(opcode):
        opcode = Opcode.JMP

    from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo
    opcode_info = OpcodeInfo.get_info(opcode)
    arg_size = opcode_info.data_len
    jmp_arg = Integer(arg_size + integer + jump_through).to_byte_array(min_length=arg_size) if integer > 0 \
        else Integer(integer).to_byte_array(min_length=arg_size)

    if len(jmp_arg) > opcode_info.max_data_len and has_larger_opcode(opcode):
        opcode = get_larger_opcode(opcode)
        opcode_info = OpcodeInfo.get_info(opcode)
        arg_size = opcode_info.data_len
        jmp_arg = Integer(arg_size + integer + jump_through).to_byte_array(min_length=arg_size) if integer > 0 \
            else Integer(integer).to_byte_array(min_length=arg_size)
        jmp_opcode = opcode
    else:
        jmp_opcode = opcode

    jmp_arg = jmp_arg[-arg_size:]

    return jmp_opcode, jmp_arg


def has_target(opcode: Opcode) -> bool:
    return Opcode.JMP <= opcode <= Opcode.CALL_L or Opcode.TRY <= opcode < Opcode.ENDFINALLY


def get_drop(position: int) -> Optional[Opcode]:
    """
    Gets the opcode to remove the item n back in the stack

    :param position: index of the variable
    :return: the respective opcode
    :rtype: Opcode
    """
    duplicate_item = {
        1: Opcode.DROP,
        2: Opcode.NIP
    }

    if position > 0:
        if position in duplicate_item:
            return duplicate_item[position]
        else:
            return Opcode.XDROP


def get_dup(position: int) -> Optional[Opcode]:
    """
    Gets the opcode to duplicate the item n back in the stack

    :param position: index of the variable
    :return: the respective opcode
    :rtype: Opcode
    """
    duplicate_item = {
        1: Opcode.DUP,
        2: Opcode.OVER
    }

    if position >= 0:
        if position in duplicate_item:
            return duplicate_item[position]
        else:
            return Opcode.PICK


def get_reverse(no_stack_items: int) -> Optional[Opcode]:
    """
    Gets the opcode to reverse n items on the stack

    :param no_stack_items: index of the variable
    :return: the respective opcode
    :rtype: Opcode
    """
    reverse_stack = {
        2: Opcode.SWAP,
        3: Opcode.REVERSE3,
        4: Opcode.REVERSE4
    }

    if no_stack_items >= 0:
        if no_stack_items in reverse_stack:
            return reverse_stack[no_stack_items]
        else:
            return Opcode.REVERSEN


def get_store(index: int, local: bool, is_arg: bool = False) -> Opcode:
    """
    Gets the opcode to store the variable

    :param index: index of the variable
    :param local: identifies if the variable is local or global
    :param is_arg: identifies if the variable is an argument of a function. False if local is False.
    :return: the respective opcode
    :rtype: Opcode
    """
    if not local:
        is_arg = False

    if 0 <= index <= 6:
        if is_arg:
            opcode_value: int = Integer.from_bytes(Opcode.STARG0) + index
        elif local:
            opcode_value: int = Integer.from_bytes(Opcode.STLOC0) + index
        else:
            opcode_value: int = Integer.from_bytes(Opcode.STSFLD0) + index
        return Opcode(Integer(opcode_value).to_byte_array())
    else:
        if is_arg:
            return Opcode.STARG
        elif local:
            return Opcode.STLOC
        else:
            return Opcode.STSFLD


def get_load(index: int, local: bool, is_arg: bool = False) -> Opcode:
    """
    Gets the opcode to load the variable

    :param index: index of the variable
    :param local: identifies if the variable is local or global
    :param is_arg: identifies if the variable is an argument of a function. False if local is False.
    :return: the respective opcode
    :rtype: Opcode
    """
    if not local:
        is_arg = False

    if 0 <= index <= 6:
        if is_arg:
            opcode_value: int = Integer.from_bytes(Opcode.LDARG0) + index
        elif local:
            opcode_value: int = Integer.from_bytes(Opcode.LDLOC0) + index
        else:
            opcode_value: int = Integer.from_bytes(Opcode.LDSFLD0) + index
        return Opcode(Integer(opcode_value).to_byte_array())
    else:
        if is_arg:
            return Opcode.LDARG
        elif local:
            return Opcode.LDLOC
        else:
            return Opcode.LDSFLD


def is_load_slot(opcode: Opcode) -> bool:
    return (Opcode.LDSFLD0 <= opcode <= Opcode.LDSFLD
            or Opcode.LDLOC0 <= opcode <= Opcode.LDLOC
            or Opcode.LDARG0 <= opcode <= Opcode.LDARG)


def get_store_from_load(load_opcode) -> Optional[Opcode]:
    """
    Gets the store slot opcode equivalent to the given load slot opcode.

    :param load_opcode: load opcode
    :type load_opcode: Opcode
    :return: equivalent store opcode if the given opcode is a load slot. Otherwise, returns None
    :rtype: Opcode or None
    """
    if is_load_slot(load_opcode):
        opcode_value: int = Integer.from_bytes(load_opcode) + 8
        return Opcode(Integer(opcode_value).to_byte_array())



from __future__ import annotations

import ast
from abc import ABC
from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ContractGetHashMethod(IBuiltinMethod, ABC):
    def __init__(self, script_hash: bytes, identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None):
        self._script_hash = script_hash
        super().__init__(identifier, args, defaults, return_type, vararg, kwargs)

    @property
    def script_hash(self) -> bytes:
        return self._script_hash

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer

        value = self.script_hash
        return [
            (Opcode.PUSHDATA1, Integer(len(value)).to_byte_array() + value)
        ]



from typing import Dict, List, Tuple

from boa3.model.builtin.method.strmethod import StrMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String


class StrBoolMethod(StrMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        args: Dict[str, Variable] = {
            'object': Variable(Type.bool),
        }

        super().__init__(args)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')
        true = String("True").to_bytes()
        false = String("False").to_bytes()

        check_is_true = [
            jmp_place_holder
        ]

        put_false = [
            (Opcode.PUSHDATA1, Integer(len(false)).to_byte_array(signed=True, min_length=1) + false),
            jmp_place_holder
        ]

        put_true = [
            (Opcode.PUSHDATA1, Integer(len(true)).to_byte_array(signed=True, min_length=1) + true),
        ]

        jmp_put_true = OpcodeHelper.get_jump_and_data(Opcode.JMP, get_bytes_count(put_true))
        put_false[-1] = jmp_put_true

        jmp_put_false = OpcodeHelper.get_jump_and_data(Opcode.JMPIF, get_bytes_count(put_false))
        check_is_true[-1] = jmp_put_false

        return (
            check_is_true +
            put_false +
            put_true
        )



from typing import Dict

from boa3.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.model.variable import Variable


class Murmur32Method(CryptoLibMethod):
    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.primitive.bytestringtype import ByteStringType
        identifier = 'murmur32'
        native_identifier = 'murmur32'
        args: Dict[str, Variable] = {
            'data': Variable(ByteStringType.build()),
            'seed': Variable(Type.int),
        }
        super().__init__(identifier, native_identifier, args, return_type=ByteStringType.build())



from boa3.builtin.compile_time import public


@public
def main() -> bytes:
    c = b'\x01\x04\xaf\x09'

    l = len(c)

    b = c[2:1]

    j = b'\x01\x02\x03\x04\x05\x06\x07'

    k = c + j

    m = k[3:6]

    return m + b



from __future__ import annotations

from typing import Any, Dict

from boa3.model.builtin.interop.nativecontract import OracleContract
from boa3.model.builtin.native.inativecontractclass import INativeContractClass
from boa3.model.method import Method


class OracleClass(INativeContractClass):
    """
    A class used to represent Oracle class
    """

    def __init__(self):
        super().__init__('Oracle', OracleContract)

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.interop.oracle.oraclegetpricemethod import OracleGetPriceMethod
        from boa3.model.builtin.interop.oracle.oraclerequestmethod import OracleRequestMethod

        if len(self._class_methods) == 0:
            self._class_methods = {
                'get_price': OracleGetPriceMethod(),
                'request': OracleRequestMethod()
            }
        return super().class_methods

    @classmethod
    def build(cls, value: Any = None) -> OracleClass:
        if value is None or cls._is_type_of(value):
            return _Oracle

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, OracleClass)


_Oracle = OracleClass()



from typing import Any, List

from boa3.builtin.interop.crypto.namedcurve import NamedCurve
from boa3.builtin.type import ByteString, ECPoint


def sha256(key: Any) -> bytes:
    """
    Encrypts a key using SHA-256.

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def ripemd160(key: Any) -> bytes:
    """
    Encrypts a key using RIPEMD-160.

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def hash160(key: Any) -> bytes:
    """
    Encrypts a key using HASH160.

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def hash256(key: Any) -> bytes:
    """
    Encrypts a key using HASH256.

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def check_sig(pub_key: ECPoint, signature: bytes) -> bool:
    """
    Checks the signature for the current script container.

    :param pub_key: the public key of the account
    :type pub_key: ECPoint
    :param signature: the signature of the current script container
    :type signature: bytes
    :return: whether the signature is valid or not
    :rtype: bool
    """
    pass


def check_multisig(pubkeys: List[ECPoint], signatures: List[bytes]) -> bool:
    """
    Checks the signatures for the current script container.

    :param pubkeys: a list of public keys
    :type pubkeys: List[ECPoint]
    :param signatures: a list of signatures
    :type signatures: List[bytes]
    :return: a boolean value that represents whether the signatures were validated
    :rtype: bool
    """
    pass


def verify_with_ecdsa(message: Any, pubkey: ECPoint, signature: ByteString, curve: NamedCurve) -> bool:
    """
    Using the elliptic curve, it checks if the signature of the any item was originally produced by the public key.

    :param message: the encrypted message
    :type message: Any
    :param pubkey: the public key that might have created the item
    :type pubkey: ECPoint
    :param signature: the signature of the item
    :type signature: bytes
    :param curve: the curve that will be used by the ecdsa
    :type curve: NamedCurve
    :return: a boolean value that represents whether the signature is valid
    :rtype: bool
    """
    pass


def murmur32(data: ByteString, seed: int) -> ByteString:
    """
    Computes the hash value for the specified byte array using the murmur32 algorithm.

    :param data: the input to compute the hash code for
    :type data: ByteString
    :param seed: the seed of the murmur32 hash function
    :type seed: int
    :return: the hash value
    :rtype: ByteString
    """
    pass



import ast
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.primitive.ibytestringtype import IByteStringType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class StripMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        from boa3.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            from boa3.model.type.primitive.bytestringtype import ByteStringType
            self_type = ByteStringType.build()

        identifier = 'strip'
        args: Dict[str, Variable] = {
            'self': Variable(self_type),
            'chars': Variable(self_type),
        }

        from string import whitespace

        # whitespace is ' \t\n\r\v\f', but it needs to be r' \t\n\r\v\f'
        # encode('unicode-escape') will make it a bytes equivalent of r' \t\n\r\v\f' and
        # decode() will convert it back to str
        whitespace_chars = [char.encode('unicode-escape').decode() for char in whitespace]

        if Type.str.is_type_of(self_type):
            chars_default = ast.parse("'{0}'".format(''.join(whitespace_chars))).body[0].value
        else:
            chars_default = ast.parse("b'{0}'".format(''.join(whitespace_chars))).body[0].value

        super().__init__(identifier, args, defaults=[chars_default], return_type=self_type)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.StackItem import StackItemType

        jmp_place_holder = (Opcode.JMP, b'\x01')

        # receive: string, chars

        initializing_first_loop = [          # initializes the variables for the first loop
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),              # string_size = len(string)
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),
            (Opcode.PUSH0, b''),             # index = 0
        ]

        verify_leading_chars = [             # verifies if all leading characters are in chars
            (Opcode.DUP, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            jmp_place_holder,                # if index >= string_size, jump to initialize_second_loop
        ]

        get_leading_char_at_index = [        # gets the character at the current index and create another index variable
            (Opcode.PUSH4, b''),             # to go through the chars
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PUSH1, b''),
            (Opcode.SUBSTR, b''),            # char = string[index]
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.PUSH0, b''),             # index_chars = 0
        ]

        verify_if_index_gt_len_chars = [     # verify if already compared all chars with char
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            jmp_place_holder,                # if index >= string_size, jump to remove_extras
        ]

        verify_if_char_in_chars = [  # compares char with chars[index_chars]
            (Opcode.PUSH5, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PUSH1, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.NUMEQUAL, b''),          # equal = char == chars[index_chars]
            (Opcode.SWAP, b''),
            (Opcode.INC, b''),               # index_chars++
            (Opcode.SWAP, b''),
            # jump back to verify_if_index_gt_len_chars if not equal
        ]

        jmp_back_to_verify_index = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT,
                                                                  -get_bytes_count(verify_if_char_in_chars +
                                                                                   verify_if_index_gt_len_chars))
        verify_if_char_in_chars.append(jmp_back_to_verify_index)

        leading_char_found = [               # char is in chars, so go back to verify_leading_chars
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.INC, b''),               # index++
            # jump back to verify_leading_chars
        ]

        jmp_back_to_verify_leading_chars = OpcodeHelper.get_jump_and_data(Opcode.JMP,
                                                                          -get_bytes_count(verify_leading_chars +
                                                                                           get_leading_char_at_index +
                                                                                           verify_if_index_gt_len_chars +
                                                                                           verify_if_char_in_chars +
                                                                                           leading_char_found))

        leading_char_found.append(jmp_back_to_verify_leading_chars)

        jmp_to_leading_chars = OpcodeHelper.get_jump_and_data(Opcode.JMPGE, get_bytes_count(verify_if_char_in_chars +
                                                                                            leading_char_found), True)
        verify_if_index_gt_len_chars[-1] = jmp_to_leading_chars

        remove_extras = [                   # remove opcodes that won't be used anymore
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        jmp_to_leading_chars = OpcodeHelper.get_jump_and_data(Opcode.JMPGE, get_bytes_count(get_leading_char_at_index +
                                                                                            verify_if_index_gt_len_chars +
                                                                                            verify_if_char_in_chars +
                                                                                            leading_char_found +
                                                                                            remove_extras), True)
        verify_leading_chars[-1] = jmp_to_leading_chars

        get_number_of_leading_chars = (
            initializing_first_loop +
            verify_leading_chars +
            get_leading_char_at_index +
            verify_if_index_gt_len_chars +
            verify_if_char_in_chars +
            leading_char_found +
            remove_extras
        )

        initialize_second_loop = [          # initializes the variables for the second loop
            (Opcode.REVERSE3, b''),         # n_leading = index
            (Opcode.DEC, b''),              # index = string_size - 1
        ]

        verify_trailing_chars = [           # verifies if leading and trailing characters are the same
            (Opcode.DUP, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            jmp_place_holder,               # if index <= n_leading, go strip the string
        ]

        get_trailing_char_at_index = [      # gets the character at the current index and create another index variable
            (Opcode.PUSH4, b''),            # to go through the chars
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PUSH1, b''),
            (Opcode.SUBSTR, b''),           # char = string[index]
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.PUSH0, b''),            # index_chars = 0
        ]

        get_trailing_char_at_index.extend(verify_if_index_gt_len_chars)
        get_trailing_char_at_index.extend(verify_if_char_in_chars)

        trailing_char_found = [             # char is in chars, so go back to get_trailing_char_at_index
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DEC, b''),              # index--
            # jump back to get_trailing_char_at_index
        ]

        jmp_back_to_verify_trailing_chars = OpcodeHelper.get_jump_and_data(Opcode.JMP,
                                                                           -get_bytes_count(get_trailing_char_at_index +
                                                                                            trailing_char_found))

        trailing_char_found.append(jmp_back_to_verify_trailing_chars)

        remove_extras = [                   # remove opcodes that won't be used anymore
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        jmp_to_strip_string = OpcodeHelper.get_jump_and_data(Opcode.JMPLE, get_bytes_count(get_trailing_char_at_index +
                                                                                           trailing_char_found +
                                                                                           remove_extras), True)
        verify_trailing_chars[-1] = jmp_to_strip_string

        get_number_of_trailing_chars = (
            initialize_second_loop +
            verify_trailing_chars +
            get_trailing_char_at_index +
            trailing_char_found +
            remove_extras
        )

        strip_string = [                    # strips the string using chars
            (Opcode.NIP, b''),
            (Opcode.OVER, b''),
            (Opcode.SUB, b''),
            (Opcode.INC, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            (Opcode.SWAP, b''),
            (Opcode.SUBSTR, b''),           # string.strip(chars)
            (Opcode.CONVERT, StackItemType.ByteString),
        ]

        return (
            get_number_of_leading_chars +
            get_number_of_trailing_chars +
            strip_string
        )

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, IByteStringType):
            return StripMethod(value)
        return super().build(value)
